# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/status.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%sarus_data_spec/protobuf/status.proto\x12\x0fsarus_data_spec\"\xe5\x05\n\x06Status\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61taspec\x18\x02 \x01(\t\x12\x0f\n\x07manager\x18\x03 \x01(\t\x12\x10\n\x08\x64\x61tetime\x18\x04 \x01(\t\x12<\n\x0btask_stages\x18\x05 \x03(\x0b\x32\'.sarus_data_spec.Status.TaskStagesEntry\x12;\n\nproperties\x18\x06 \x03(\x0b\x32\'.sarus_data_spec.Status.PropertiesEntry\x1aP\n\x0fTaskStagesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x1d.sarus_data_spec.Status.Stage:\x02\x38\x01\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x97\x03\n\x05Stage\x12\x38\n\x07pending\x18\x01 \x01(\x0b\x32%.sarus_data_spec.Status.Stage.PendingH\x00\x12>\n\nprocessing\x18\x02 \x01(\x0b\x32(.sarus_data_spec.Status.Stage.ProcessingH\x00\x12\x34\n\x05ready\x18\x03 \x01(\x0b\x32#.sarus_data_spec.Status.Stage.ReadyH\x00\x12\x34\n\x05\x65rror\x18\x04 \x01(\x0b\x32#.sarus_data_spec.Status.Stage.ErrorH\x00\x12\x41\n\nproperties\x18\x05 \x03(\x0b\x32-.sarus_data_spec.Status.Stage.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\t\n\x07Pending\x1a\x0c\n\nProcessing\x1a\x07\n\x05Ready\x1a\x07\n\x05\x45rrorB\x07\n\x05stageb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.status_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _STATUS_TASKSTAGESENTRY._options = None
  _STATUS_TASKSTAGESENTRY._serialized_options = b'8\001'
  _STATUS_PROPERTIESENTRY._options = None
  _STATUS_PROPERTIESENTRY._serialized_options = b'8\001'
  _STATUS_STAGE_PROPERTIESENTRY._options = None
  _STATUS_STAGE_PROPERTIESENTRY._serialized_options = b'8\001'
  _STATUS._serialized_start=59
  _STATUS._serialized_end=800
  _STATUS_TASKSTAGESENTRY._serialized_start=259
  _STATUS_TASKSTAGESENTRY._serialized_end=339
  _STATUS_PROPERTIESENTRY._serialized_start=341
  _STATUS_PROPERTIESENTRY._serialized_end=390
  _STATUS_STAGE._serialized_start=393
  _STATUS_STAGE._serialized_end=800
  _STATUS_STAGE_PROPERTIESENTRY._serialized_start=341
  _STATUS_STAGE_PROPERTIESENTRY._serialized_end=390
  _STATUS_STAGE_PENDING._serialized_start=750
  _STATUS_STAGE_PENDING._serialized_end=759
  _STATUS_STAGE_PROCESSING._serialized_start=761
  _STATUS_STAGE_PROCESSING._serialized_end=773
  _STATUS_STAGE_READY._serialized_start=775
  _STATUS_STAGE_READY._serialized_end=782
  _STATUS_STAGE_ERROR._serialized_start=784
  _STATUS_STAGE_ERROR._serialized_end=791
# @@protoc_insertion_point(module_scope)
