# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/statistics.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n)sarus_data_spec/protobuf/statistics.proto\x12\x0fsarus_data_spec\"\xbb\x18\n\nStatistics\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x30\n\x04null\x18\x03 \x01(\x0b\x32 .sarus_data_spec.Statistics.NullH\x00\x12\x30\n\x04unit\x18\x04 \x01(\x0b\x32 .sarus_data_spec.Statistics.UnitH\x00\x12\x36\n\x07\x62oolean\x18\x05 \x01(\x0b\x32#.sarus_data_spec.Statistics.BooleanH\x00\x12\x36\n\x07integer\x18\x06 \x01(\x0b\x32#.sarus_data_spec.Statistics.IntegerH\x00\x12\x30\n\x04\x65num\x18\x07 \x01(\x0b\x32 .sarus_data_spec.Statistics.EnumH\x00\x12\x32\n\x05\x66loat\x18\x08 \x01(\x0b\x32!.sarus_data_spec.Statistics.FloatH\x00\x12\x30\n\x04text\x18\t \x01(\x0b\x32 .sarus_data_spec.Statistics.TextH\x00\x12\x32\n\x05\x62ytes\x18\n \x01(\x0b\x32!.sarus_data_spec.Statistics.BytesH\x00\x12\x34\n\x06struct\x18\x0b \x01(\x0b\x32\".sarus_data_spec.Statistics.StructH\x00\x12\x32\n\x05union\x18\x0c \x01(\x0b\x32!.sarus_data_spec.Statistics.UnionH\x00\x12\x38\n\x08optional\x18\r \x01(\x0b\x32$.sarus_data_spec.Statistics.OptionalH\x00\x12\x30\n\x04list\x18\x0e \x01(\x0b\x32 .sarus_data_spec.Statistics.ListH\x00\x12\x32\n\x05\x61rray\x18\x0f \x01(\x0b\x32!.sarus_data_spec.Statistics.ArrayH\x00\x12\x38\n\x08\x64\x61tetime\x18\x10 \x01(\x0b\x32$.sarus_data_spec.Statistics.DatetimeH\x00\x12,\n\x02id\x18\x11 \x01(\x0b\x32\x1e.sarus_data_spec.Statistics.IdH\x00\x12>\n\x0b\x63onstrained\x18\x12 \x01(\x0b\x32\'.sarus_data_spec.Statistics.ConstrainedH\x00\x12\x30\n\x04\x64\x61te\x18\x13 \x01(\x0b\x32 .sarus_data_spec.Statistics.DateH\x00\x12\x30\n\x04time\x18\x14 \x01(\x0b\x32 .sarus_data_spec.Statistics.TimeH\x00\x12\x38\n\x08\x64uration\x18\x15 \x01(\x0b\x32$.sarus_data_spec.Statistics.DurationH\x00\x12?\n\nproperties\x18\x02 \x03(\x0b\x32+.sarus_data_spec.Statistics.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a*\n\x04Null\x12\x0c\n\x04size\x18\x01 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x02 \x01(\x01\x1a*\n\x04Unit\x12\x0c\n\x04size\x18\x01 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x02 \x01(\x01\x1a\x62\n\x07\x42oolean\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1a(\n\x02Id\x12\x0c\n\x04size\x18\x01 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x02 \x01(\x01\x1a\x62\n\x07Integer\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1a_\n\x04\x45num\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1a`\n\x05\x46loat\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1ap\n\x04Text\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0f\n\x07\x65xample\x18\x02 \x01(\t\x12\x0c\n\x04size\x18\x03 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x04 \x01(\x01\x1a+\n\x05\x42ytes\x12\x0c\n\x04size\x18\x01 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x02 \x01(\x01\x1a\xbc\x01\n\x06Struct\x12\x38\n\x06\x66ields\x18\x01 \x03(\x0b\x32(.sarus_data_spec.Statistics.Struct.Field\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x14\n\x0cmultiplicity\x18\x04 \x01(\x01\x1a\x46\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\nstatistics\x18\x02 \x01(\x0b\x32\x1b.sarus_data_spec.Statistics\x1a\xba\x01\n\x05Union\x12\x37\n\x06\x66ields\x18\x01 \x03(\x0b\x32\'.sarus_data_spec.Statistics.Union.Field\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x14\n\x0cmultiplicity\x18\x04 \x01(\x01\x1a\x46\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\nstatistics\x18\x02 \x01(\x0b\x32\x1b.sarus_data_spec.Statistics\x1am\n\x08Optional\x12/\n\nstatistics\x18\x01 \x01(\x0b\x32\x1b.sarus_data_spec.Statistics\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x14\n\x0cmultiplicity\x18\x04 \x01(\x01\x1a\x90\x01\n\x04List\x12/\n\nstatistics\x18\x01 \x01(\x0b\x32\x1b.sarus_data_spec.Statistics\x12\x33\n\x0c\x64istribution\x18\x02 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x03 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x04 \x01(\x01\x1a\x92\x01\n\x05\x41rray\x12/\n\nstatistics\x18\x01 \x01(\x0b\x32\x1b.sarus_data_spec.Statistics\x12\x34\n\rdistributions\x18\x02 \x03(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x03 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x04 \x01(\x01\x1a\x63\n\x08\x44\x61tetime\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1a_\n\x04\x44\x61te\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1a_\n\x04Time\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1a\x63\n\x08\x44uration\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1a\x62\n\x0b\x43onstrained\x12/\n\nstatistics\x18\x01 \x01(\x0b\x32\x1b.sarus_data_spec.Statistics\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x42\x0c\n\nstatistics\"\x93\x07\n\x0c\x44istribution\x12\x36\n\x06\x64ouble\x18\x01 \x01(\x0b\x32$.sarus_data_spec.Distribution.DoubleH\x00\x12\x38\n\x07integer\x18\x02 \x01(\x0b\x32%.sarus_data_spec.Distribution.IntegerH\x00\x12\x38\n\x07\x62oolean\x18\x03 \x01(\x0b\x32%.sarus_data_spec.Distribution.BooleanH\x00\x12\x32\n\x04\x65num\x18\x04 \x01(\x0b\x32\".sarus_data_spec.Distribution.EnumH\x00\x12\x41\n\nproperties\x18\x05 \x03(\x0b\x32-.sarus_data_spec.Distribution.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x8b\x01\n\x06\x44ouble\x12:\n\x06points\x18\x01 \x03(\x0b\x32*.sarus_data_spec.Distribution.Double.Point\x12\x0b\n\x03min\x18\x02 \x01(\x01\x12\x0b\n\x03max\x18\x03 \x01(\x01\x1a+\n\x05Point\x12\r\n\x05value\x18\x02 \x01(\x01\x12\x13\n\x0bprobability\x18\x03 \x01(\x01\x1a\x8d\x01\n\x07Integer\x12;\n\x06points\x18\x01 \x03(\x0b\x32+.sarus_data_spec.Distribution.Integer.Point\x12\x0b\n\x03min\x18\x02 \x01(\x03\x12\x0b\n\x03max\x18\x03 \x01(\x03\x1a+\n\x05Point\x12\r\n\x05value\x18\x01 \x01(\x03\x12\x13\n\x0bprobability\x18\x02 \x01(\x01\x1a\x81\x01\n\x07\x42oolean\x12;\n\x06points\x18\x01 \x03(\x0b\x32+.sarus_data_spec.Distribution.Boolean.Point\x1a\x39\n\x05Point\x12\x0c\n\x04name\x18\x01 \x01(\x08\x12\r\n\x05value\x18\x02 \x01(\x03\x12\x13\n\x0bprobability\x18\x03 \x01(\x01\x1a{\n\x04\x45num\x12\x38\n\x06points\x18\x01 \x03(\x0b\x32(.sarus_data_spec.Distribution.Enum.Point\x1a\x39\n\x05Point\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x01\x12\x13\n\x0bprobability\x18\x03 \x01(\x01\x42\x0e\n\x0c\x64istributionb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.statistics_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _STATISTICS_PROPERTIESENTRY._options = None
  _STATISTICS_PROPERTIESENTRY._serialized_options = b'8\001'
  _DISTRIBUTION_PROPERTIESENTRY._options = None
  _DISTRIBUTION_PROPERTIESENTRY._serialized_options = b'8\001'
  _STATISTICS._serialized_start=63
  _STATISTICS._serialized_end=3194
  _STATISTICS_PROPERTIESENTRY._serialized_start=1164
  _STATISTICS_PROPERTIESENTRY._serialized_end=1213
  _STATISTICS_NULL._serialized_start=1215
  _STATISTICS_NULL._serialized_end=1257
  _STATISTICS_UNIT._serialized_start=1259
  _STATISTICS_UNIT._serialized_end=1301
  _STATISTICS_BOOLEAN._serialized_start=1303
  _STATISTICS_BOOLEAN._serialized_end=1401
  _STATISTICS_ID._serialized_start=1403
  _STATISTICS_ID._serialized_end=1443
  _STATISTICS_INTEGER._serialized_start=1445
  _STATISTICS_INTEGER._serialized_end=1543
  _STATISTICS_ENUM._serialized_start=1545
  _STATISTICS_ENUM._serialized_end=1640
  _STATISTICS_FLOAT._serialized_start=1642
  _STATISTICS_FLOAT._serialized_end=1738
  _STATISTICS_TEXT._serialized_start=1740
  _STATISTICS_TEXT._serialized_end=1852
  _STATISTICS_BYTES._serialized_start=1854
  _STATISTICS_BYTES._serialized_end=1897
  _STATISTICS_STRUCT._serialized_start=1900
  _STATISTICS_STRUCT._serialized_end=2088
  _STATISTICS_STRUCT_FIELD._serialized_start=2018
  _STATISTICS_STRUCT_FIELD._serialized_end=2088
  _STATISTICS_UNION._serialized_start=2091
  _STATISTICS_UNION._serialized_end=2277
  _STATISTICS_UNION_FIELD._serialized_start=2018
  _STATISTICS_UNION_FIELD._serialized_end=2088
  _STATISTICS_OPTIONAL._serialized_start=2279
  _STATISTICS_OPTIONAL._serialized_end=2388
  _STATISTICS_LIST._serialized_start=2391
  _STATISTICS_LIST._serialized_end=2535
  _STATISTICS_ARRAY._serialized_start=2538
  _STATISTICS_ARRAY._serialized_end=2684
  _STATISTICS_DATETIME._serialized_start=2686
  _STATISTICS_DATETIME._serialized_end=2785
  _STATISTICS_DATE._serialized_start=2787
  _STATISTICS_DATE._serialized_end=2882
  _STATISTICS_TIME._serialized_start=2884
  _STATISTICS_TIME._serialized_end=2979
  _STATISTICS_DURATION._serialized_start=2981
  _STATISTICS_DURATION._serialized_end=3080
  _STATISTICS_CONSTRAINED._serialized_start=3082
  _STATISTICS_CONSTRAINED._serialized_end=3180
  _DISTRIBUTION._serialized_start=3197
  _DISTRIBUTION._serialized_end=4112
  _DISTRIBUTION_PROPERTIESENTRY._serialized_start=1164
  _DISTRIBUTION_PROPERTIESENTRY._serialized_end=1213
  _DISTRIBUTION_DOUBLE._serialized_start=3556
  _DISTRIBUTION_DOUBLE._serialized_end=3695
  _DISTRIBUTION_DOUBLE_POINT._serialized_start=3652
  _DISTRIBUTION_DOUBLE_POINT._serialized_end=3695
  _DISTRIBUTION_INTEGER._serialized_start=3698
  _DISTRIBUTION_INTEGER._serialized_end=3839
  _DISTRIBUTION_INTEGER_POINT._serialized_start=3796
  _DISTRIBUTION_INTEGER_POINT._serialized_end=3839
  _DISTRIBUTION_BOOLEAN._serialized_start=3842
  _DISTRIBUTION_BOOLEAN._serialized_end=3971
  _DISTRIBUTION_BOOLEAN_POINT._serialized_start=3914
  _DISTRIBUTION_BOOLEAN_POINT._serialized_end=3971
  _DISTRIBUTION_ENUM._serialized_start=3973
  _DISTRIBUTION_ENUM._serialized_end=4096
  _DISTRIBUTION_ENUM_POINT._serialized_start=4039
  _DISTRIBUTION_ENUM_POINT._serialized_end=4096
# @@protoc_insertion_point(module_scope)
