# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/links.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from sarus_data_spec.protobuf import path_pb2 as sarus__data__spec_dot_protobuf_dot_path__pb2
from sarus_data_spec.protobuf import statistics_pb2 as sarus__data__spec_dot_protobuf_dot_statistics__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$sarus_data_spec/protobuf/links.proto\x12\x0fsarus_data_spec\x1a#sarus_data_spec/protobuf/path.proto\x1a)sarus_data_spec/protobuf/statistics.proto\"\xdb\x03\n\x05Links\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0f\n\x07\x64\x61taset\x18\x02 \x01(\t\x12\x39\n\x10links_statistics\x18\x03 \x03(\x0b\x32\x1f.sarus_data_spec.Links.LinkStat\x12:\n\nproperties\x18\x04 \x03(\x0b\x32&.sarus_data_spec.Links.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x88\x02\n\x08LinkStat\x12\'\n\x08pointing\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Path\x12&\n\x07pointed\x18\x02 \x01(\x0b\x32\x15.sarus_data_spec.Path\x12\x33\n\x0c\x64istribution\x18\x03 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x43\n\nproperties\x18\x04 \x03(\x0b\x32/.sarus_data_spec.Links.LinkStat.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.links_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _LINKS_PROPERTIESENTRY._options = None
  _LINKS_PROPERTIESENTRY._serialized_options = b'8\001'
  _LINKS_LINKSTAT_PROPERTIESENTRY._options = None
  _LINKS_LINKSTAT_PROPERTIESENTRY._serialized_options = b'8\001'
  _LINKS._serialized_start=138
  _LINKS._serialized_end=613
  _LINKS_PROPERTIESENTRY._serialized_start=297
  _LINKS_PROPERTIESENTRY._serialized_end=346
  _LINKS_LINKSTAT._serialized_start=349
  _LINKS_LINKSTAT._serialized_end=613
  _LINKS_LINKSTAT_PROPERTIESENTRY._serialized_start=297
  _LINKS_LINKSTAT_PROPERTIESENTRY._serialized_end=346
# @@protoc_insertion_point(module_scope)
