# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/relation.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'sarus_data_spec/protobuf/relation.proto\x12\x0fsarus_data_spec\"\xc4\x01\n\x08Relation\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x11\n\targuments\x18\x03 \x03(\t\x12\x17\n\x0fnamed_arguments\x18\x04 \x03(\t\x12=\n\nproperties\x18\x05 \x03(\x0b\x32).sarus_data_spec.Relation.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.relation_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _RELATION_PROPERTIESENTRY._options = None
  _RELATION_PROPERTIESENTRY._serialized_options = b'8\001'
  _RELATION._serialized_start=61
  _RELATION._serialized_end=257
  _RELATION_PROPERTIESENTRY._serialized_start=208
  _RELATION_PROPERTIESENTRY._serialized_end=257
# @@protoc_insertion_point(module_scope)
