syntax = "proto3";
package sarus_data_spec;

message Attribute {
  // General attribute about an object
  string uuid = 1;
  string object = 2; // uuid of an object
  map<string, string> properties = 3; // Other properties
  string name = 4;
}
