# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/predicate.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(sarus_data_spec/protobuf/predicate.proto\x12\x0fsarus_data_spec\"\xab\x04\n\tPredicate\x12\x33\n\x06simple\x18\x01 \x01(\x0b\x32!.sarus_data_spec.Predicate.SimpleH\x00\x12\x31\n\x05inter\x18\x02 \x01(\x0b\x32 .sarus_data_spec.Predicate.InterH\x00\x12\x31\n\x05union\x18\x03 \x01(\x0b\x32 .sarus_data_spec.Predicate.UnionH\x00\x12/\n\x04\x63omp\x18\x04 \x01(\x0b\x32\x1f.sarus_data_spec.Predicate.CompH\x00\x12>\n\nproperties\x18\x05 \x03(\x0b\x32*.sarus_data_spec.Predicate.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a)\n\x06Simple\x12\x10\n\x08operator\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\x1a\x37\n\x05Inter\x12.\n\npredicates\x18\x01 \x03(\x0b\x32\x1a.sarus_data_spec.Predicate\x1a\x37\n\x05Union\x12.\n\npredicates\x18\x01 \x03(\x0b\x32\x1a.sarus_data_spec.Predicate\x1a\x35\n\x04\x43omp\x12-\n\tpredicate\x18\x01 \x01(\x0b\x32\x1a.sarus_data_spec.PredicateB\x0b\n\tpredicateb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.predicate_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _PREDICATE_PROPERTIESENTRY._options = None
  _PREDICATE_PROPERTIESENTRY._serialized_options = b'8\001'
  _PREDICATE._serialized_start=62
  _PREDICATE._serialized_end=617
  _PREDICATE_PROPERTIESENTRY._serialized_start=343
  _PREDICATE_PROPERTIESENTRY._serialized_end=392
  _PREDICATE_SIMPLE._serialized_start=394
  _PREDICATE_SIMPLE._serialized_end=435
  _PREDICATE_INTER._serialized_start=437
  _PREDICATE_INTER._serialized_end=492
  _PREDICATE_UNION._serialized_start=494
  _PREDICATE_UNION._serialized_end=549
  _PREDICATE_COMP._serialized_start=551
  _PREDICATE_COMP._serialized_end=604
# @@protoc_insertion_point(module_scope)
